import inspect
import os

import mock

from .base import TestCase

from doctor.utils import (
    exec_params, get_description_lines, get_module_attr, nested_set)


class TestUtils(TestCase):

    def test_exec_params_function(self):
        def logic(a, b, c=None):
            return a+1, b+1, c
        logic._argspec = inspect.getargspec(logic)
        kwargs = {'c': 1, 'd': 'a'}
        args = (1, 2)
        a, b, c = exec_params(logic, *args, **kwargs)
        self.assertEqual(a, args[0]+1)
        self.assertEqual(b, args[1]+1)
        self.assertEqual(c, kwargs['c'])

    def test_exec_params_callable(self):
        kwargs = {'c': 1, 'd': 'a'}
        args = (1, 2)

        class Foo(object):
            def __call__(self, a, b, c=None):
                return a+1, b+1, c
        logic = Foo()
        logic._argspec = inspect.getargspec(logic.__call__)
        a, b, c = exec_params(logic, *args, **kwargs)
        self.assertEqual(a, args[0]+1)
        self.assertEqual(b, args[1]+1)
        self.assertEqual(c, kwargs['c'])

    def test_exec_params_args_only(self):
        def logic(a, b, c):
            return a+1, b+1, c
        logic._argspec = inspect.getargspec(logic)
        kwargs = {'d': 1, 'e': 2}
        args = (1, 2, 3)
        a, b, c = exec_params(logic, *args, **kwargs)
        self.assertEqual(a, args[0]+1)
        self.assertEqual(b, args[1]+1)
        self.assertEqual(c, 3)

    def test_exec_params_kwargs_only(self):
        def logic(a=1, b=2, c=3):
            return a+1, b+1, c
        logic._argspec = inspect.getargspec(logic)
        kwargs = {'d': 1, 'e': 2}
        args = (1, 2, 3)
        a, b, c = exec_params(logic, *args, **kwargs)
        self.assertEqual(a, args[0]+1)
        self.assertEqual(b, args[1]+1)
        self.assertEqual(c, 3)

    @mock.patch('doctor.utils.execfile', create=True)
    @mock.patch('doctor.utils.os', autospec=True)
    @mock.patch('doctor.utils.sys', autospec=True)
    def test_get_module_attr(self, mock_sys, mock_os, mock_execfile):
        def side_effect(module_filename, namespace):
            self.assertEqual(mock_sys.path, ['foo', 'bar', '/foo/bar'])
            self.assertEqual(module_filename, '/foo/bar/baz')
            self.assertEqual(namespace, {'__file__': module_filename})
            namespace['mock_attr'] = mock.sentinel.mock_attr
        mock_execfile.side_effect = side_effect
        mock_os.getcwd.return_value = mock.sentinel.old_cwd
        mock_os.path = os.path
        mock_sys.path = ['foo', 'bar']
        result = get_module_attr('/foo/bar/baz', 'mock_attr')
        self.assertEqual(result, mock.sentinel.mock_attr)
        self.assertEqual(len(mock_execfile.call_args_list), 1)
        self.assertEqual(mock_os.chdir.call_args_list,
                         [mock.call('/foo/bar'),
                          mock.call(mock.sentinel.old_cwd)])
        self.assertEqual(mock_sys.path, ['foo', 'bar'])

    def test_get_description_lines(self):
        """
        Tests that get_description_lines properly dedents docstrings and strips
        out values we don't want.
        """
        docstring = """line one

                       line two

                           indented line three

                       line four

                       :param str c: example param
                       :returns: d
                       """
        self.assertEqual(get_description_lines(docstring), [
            'line one',
            '',
            'line two',
            '',
            '    indented line three',
            '',
            'line four',
            '',
        ])

    def test_get_description_lines_none(self):
        """It should just return an empty list for None."""
        self.assertEqual(get_description_lines(None), [])

    def test_get_description_lines_trailing_newline(self):
        """It should add a trailing line if necessary."""
        self.assertEqual(get_description_lines('foo\n:arg'), ['foo', ''])

    def test_nested_set(self):
        data_dict = {}
        nested_set(data_dict, ["b", "v", "new"], {})
        self.assertEqual(data_dict['b']['v']['new'], {})

        nested_set(data_dict, ["b", "v", "new"], '4')
        self.assertEqual(data_dict['b']['v']['new'], '4')

        nested_set(data_dict, ["a", "v", "new"], '1')
        self.assertEqual(data_dict['a']['v']['new'], '1')

        nested_set(data_dict, ["a", "new"], 'new')
        self.assertEqual(data_dict['a']['new'], 'new')



import functools
import inspect

import jsonschema

from .parsers import parse_value
from .schema import Schema


class ResourceSchemaAnnotation(object):

    """Metadata about the schema used for a given request method.

    An instance of this class is attached to each handler method in a
    _schema_annotation attribute. It can be used for introspection about the
    schemas, to generate things like API documentation and hyper schemas from
    the code.

    :param func logic: Logic function which will handle the request.
    :param str http_method: The HTTP request method for this request (e.g. GET).
    :param doctor.resource.ResourceSchema schema: The resource schema
        object for this handler.
    :param dict request_schema: The schema used to validate the request.
    :param dict response_schema: The schema used to validate the response.
    :param str title: A short title for the route.  e.g. 'Create Foo' might
        be used for a POST method on a FooListHandler.
    """

    def __init__(self, logic, http_method, schema, request_schema,
                 response_schema, title=None):
        self.logic = logic
        self.http_method = http_method
        self.schema = schema
        self.request_schema = request_schema
        self.response_schema = response_schema
        self.title = title
        if title is None:
            if http_method.upper() == 'GET':
                self.title = 'Retrieve'
            elif http_method.upper() == 'POST':
                self.title = 'Create'
            elif http_method.upper() == 'PUT':
                self.title = 'Update'
            elif http_method.upper() == 'DELETE':
                self.title = 'Delete'

    @classmethod
    def get_annotation(cls, fn):
        """Find the _schema_annotation attribute for the given function.

        This will descend through decorators until it finds something that has
        the attribute. If it doesn't find it anywhere, it will return None.

        :param func fn: Find the attribute on this function.
        :returns: an instance of
            :class:`~doctor.resource.ResourceSchemaAnnotation` or
            None.
        """
        while fn is not None:
            if hasattr(fn, '_schema_annotation'):
                return fn._schema_annotation
            fn = getattr(fn, 'im_func', fn)
            closure = getattr(fn, '__closure__', None)
            fn = closure[0].cell_contents if closure is not None else None
        return None


class ResourceSchema(Schema):

    """
    This class extends :class:`~doctor.schema.Schema` with methods
    for generating HTTP handler functions that automatically parse and
    validate the request and response objects with a given schema.
    """

    def __init__(self, schema, handle_http=None,
                 raise_response_validation_errors=False, **kwargs):
        """
        :param dict schema: The JSON schema to use for this resource.
        :param function handle_http: The HTTP handler function that should be
            used to wrap the logic functions. You would normally pass
            :func:`doctor.flask.handle_http`.
        :param bool raise_response_validation_errors: True to raise errors
            for response validation exceptions, False to just log them
            and return gracefully.
        """
        super(ResourceSchema, self).__init__(schema, **kwargs)
        self.handle_http = handle_http
        self.raise_response_validation_errors = raise_response_validation_errors

    def _create_request_schema(self, params, required):
        """Create a JSON schema for a request.

        :param list params: A list of keys specifying which definitions from
            the base schema should be allowed in the request.
        :param list required: A subset of the params that the requester must
            specify in the request.
        :returns: a JSON schema dict
        """
        # We allow additional properties because the data this will validate
        # may also include kwargs passed by decorators on the handler method.
        schema = {'additionalProperties': True,
                  'definitions': self.resolve('#/definitions'),
                  'properties': {},
                  'required': required or (),
                  'type': 'object'}
        for param in params:
            schema['properties'][param] = {
                '$ref': '#/definitions/{}'.format(param)}
        return schema

    def _parse_params(self, params, request_schema):
        """Use the request schema to coerce string params.

        This is used for HTTP requests, in which the form parameters are all
        strings, but need to be converted to the appropriate types before
        validating them.

        :param dict params: The parameters specified in the request.
        :param dict request_schema: The JSON schema for the request.
        :returns: a dict of params parsed from the input dict.
        """
        parsed_params = {}
        properties = request_schema['properties']
        for key in properties:
            if key not in params:
                continue
            prop_schema = self.resolve('#' + key, properties)
            _, value = parse_value(
                params[key], self.resolve('#type', prop_schema), key)
            parsed_params[key] = value
        return parsed_params

    def _create_http_method(self, logic, http_method, request=None,
                            response=None, params=None, required=None,
                            before=None, after=None, allowed_exceptions=None,
                            title=None):
        """Create a handler method to be used in a handler class.

        :param callable logic: The underlying function to execute with the
            parsed and validated parameters.
        :param str http_method: HTTP method this will handle.
        :param str request: A key specifying which schema definition should be
            expected in the request.
        :param str response: A key specifying which schema definition should be
            expected in the response.
        :param list(str) params: A list of keys specifying which definitions
            from the base schema should be allowed in the request.
        :param list(str) required: A subset of the params that the requester
            must specify in the request.
        :param callable before: A method to run prior to executing handle_http.
        :param callable after: A method to run after executinghandle_http.
        :param list(class) allowed_exceptions: If specified, these exception
            classes will be re-raised instead of turning them into 500 errors.
        :param str title: A short title for the route.  e.g. 'Create Foo' might
            be used for a POST method on a FooListHandler.
        :returns: a handler function
        """
        if request:
            request_schema = self.resolve('#/definitions/{}'.format(request))
            request_validator = jsonschema.Draft4Validator(
                request_schema, resolver=self.resolver,
                format_checker=jsonschema.draft4_format_checker)
        elif params:
            request_schema = self._create_request_schema(params, required)
            request_validator = jsonschema.Draft4Validator(
                request_schema, resolver=self.resolver,
                format_checker=jsonschema.draft4_format_checker)
        else:
            request_schema = request_validator = None
        if response:
            response_schema = self.resolve('#/definitions/{}'.format(response))
            response_validator = jsonschema.Draft4Validator(
                response_schema, resolver=self.resolver,
                format_checker=jsonschema.draft4_format_checker)
        else:
            response_schema = response_validator = None

        before = before if before else (lambda *args, **kwargs: None)
        after = after if after else (lambda *args, **kwargs: None)

        try:
            logic._argspec = inspect.getargspec(logic)
        except TypeError:
            logic._argspec = inspect.getargspec(logic.__call__)

        @functools.wraps(logic)
        def fn(handler, *args, **kwargs):
            before(handler, *args, **kwargs)
            result = self.handle_http(
                self, handler, args, kwargs, logic,
                request_schema, request_validator,
                response_validator, allowed_exceptions)
            after(handler, result, *args, **kwargs)
            return result

        fn.required_args = required or []
        fn._schema_annotation = ResourceSchemaAnnotation(
            fn, http_method.upper(), self, request_schema, response_schema,
            title)
        return fn

    def http_delete(self, logic, request=None, response=None, params=None,
                    required=None, before=None, after=None,
                    allowed_exceptions=None, title=None):
        """Create a handler method for a delete request.

        :see: :meth:`_create_http_method`
        """
        return self._create_http_method(
            logic, 'DELETE', request=request, response=response, params=params,
            required=required, before=before, after=after,
            allowed_exceptions=allowed_exceptions, title=title)

    def http_get(self, logic, request=None, response=None, params=None,
                 required=None, before=None, after=None,
                 allowed_exceptions=None, title=None):
        """Create a handler method for a get request.

        :see: :meth:`_create_http_method`
        """
        return self._create_http_method(
            logic, 'GET', request=request, response=response, params=params,
            required=required, before=before, after=after,
            allowed_exceptions=allowed_exceptions, title=title)

    def http_post(self, logic, request=None, response=None, params=None,
                  required=None, before=None, after=None,
                  allowed_exceptions=None, title=None):
        """Create a handler method for a post request.

        :see: :meth:`_create_http_method`
        """
        return self._create_http_method(
            logic, 'POST', request=request, response=response, params=params,
            required=required, before=before, after=after,
            allowed_exceptions=allowed_exceptions, title=title)

    def http_put(self, logic, request=None, response=None, params=None,
                 required=None, before=None, after=None,
                 allowed_exceptions=None, title=None):
        """Create a handler method for a put request.

        :see: :meth:`_create_http_method`
        """
        return self._create_http_method(
            logic, 'PUT', request=request, response=response, params=params,
            required=required, before=before, after=after,
            allowed_exceptions=allowed_exceptions, title=title)



from __future__ import absolute_import

import logging

try:
    import flask_restful
    from flask import current_app, request
    from werkzeug.exceptions import (BadRequest, Conflict, Forbidden,
                                     HTTPException, NotFound, Unauthorized,
                                     InternalServerError)
except ImportError:  # pragma: no cover
    raise ImportError('You must install flask to use the '
                      'doctor.flask module.')

from .constants import MAX_RESPONSE_LENGTH
from .errors import (ForbiddenError, ImmutableError, InvalidValueError,
                     ParseError, NotFoundError, SchemaValidationError,
                     UnauthorizedError)
from .resource import ResourceSchema
from .router import Router
from .utils import exec_params


STATUS_CODE_MAP = {
    'POST': 201,
    'DELETE': 204,
}


class SchematicHTTPException(HTTPException):

    """Schematic specific sub-class of werkzeug's BadRequest.

    Note that this adds a flask-restful specific data attribute to the class,
    as the error wouldn't render properly without it.
    """

    def __init__(self, description=None, errobj=None):
        super(SchematicHTTPException, self).__init__(description)
        self.data = {'status': self.code, 'message': description}
        self.errobj = errobj

    def __str__(self):
        return '%d: %s: %s' % (self.code, self.name, self.description)


class HTTP400Exception(SchematicHTTPException, BadRequest):
    pass


class HTTP401Exception(SchematicHTTPException, Unauthorized):
    pass


class HTTP403Exception(SchematicHTTPException, Forbidden):
    pass


class HTTP404Exception(SchematicHTTPException, NotFound):
    pass


class HTTP409Exception(SchematicHTTPException, Conflict):
    pass


class HTTP500Exception(SchematicHTTPException, InternalServerError):
    pass


def handle_http(schema, handler, args, kwargs, logic, request_schema,
                request_validator, response_validator, allowed_exceptions):
    """Handle a Flask HTTP request.

    This is invoked by the wrapper methods generated by
    :meth:`~doctor.resource.ResourceSchema._create_http_method`.
    It parses the supplied request params and validates them using the schema.
    It then passes them to the logic function, along with any kwargs passed to
    the wrapper, and validates and responds with the result.

    :param doctor.resource.ResourceSchema schema: Instance of a
        :class:`~doctor.resource.ResourceSchema` class.
    :param handler: flask_restful.Resource: An instance of a Flask Restful
        resource class.
    :param tuple args: Any positional arguments passed to the wrapper method.
    :param dict kwargs: Any keyword arguments passed to the wrapper method.
    :param callable logic: The callable to invoke to actually perform the
        business logic for this request.
    :param dict request_schema: The JSON schema for the request (or "schema" in
        JSON Hyper Schema terminology), or None.
    :param request_validator: jsonschema validator object for the request
        schema, or None.
    :param response_validator: jsonschema validator object for the response
        schema, or None.
    :param allowed_exceptions: If specified, these exception classes will be
        re-raised instead of turning them into 500 errors.
    :type allowed_exceptions: list(class) or None
    """
    try:
        if request_schema:
            # We are checking mimetype here instead of content_type because
            # mimetype is just the content-type, where as content_type can
            # contain encoding, charset, and language information.  e.g.
            # `Content-Type: application/json; charset=UTF8`
            if request.mimetype == 'application/json':
                # This is a proper typed JSON request. The parameters will be
                # encoded into the request body as a JSON blob.
                params = request.json
            else:
                # Try to parse things from normal HTTP parameters
                params = schema._parse_params(request.values, request_schema)
            params.update(**kwargs)
            if request_validator:
                schema.validate(params, request_validator)
            response = exec_params(logic, *args, **params)
        else:
            response = exec_params(logic, *args, **kwargs)
        if response_validator:
            try:
                schema.validate(response, response_validator)
            except SchemaValidationError as e:
                response_str = str(response)
                # We need a limit on the response length because logexec
                # which wraps this log has an 8k (8192 chars) limit.
                if len(response_str) > MAX_RESPONSE_LENGTH:
                    response_str = response_str[:MAX_RESPONSE_LENGTH] + '...'
                logging.warning('Response to %s %s does not validate: %s.',
                                request.method, request.path,
                                response_str, exc_info=e)
                if schema.raise_response_validation_errors:
                    raise
        return response, STATUS_CODE_MAP.get(request.method, 200)
    except (InvalidValueError, ParseError, SchemaValidationError) as e:
        errors = getattr(e, 'errors', None)
        raise HTTP400Exception(unicode(e), errobj=errors)
    except UnauthorizedError as e:
        raise HTTP401Exception(unicode(e))
    except ForbiddenError as e:
        raise HTTP403Exception(unicode(e))
    except NotFoundError as e:
        raise HTTP404Exception(unicode(e))
    except ImmutableError as e:
        raise HTTP409Exception(unicode(e))
    except Exception as e:
        # Always re-raise exceptions when DEBUG is enabled for development.
        if current_app.config.get('DEBUG', False):
            raise
        if allowed_exceptions and any(isinstance(e, cls)
                                      for cls in allowed_exceptions):
            raise
        logging.exception(unicode(e))
        raise HTTP500Exception('Uncaught error in logic function')


class FlaskResourceSchema(ResourceSchema):

    """A ResourceSchema specifically for Flask Restful resources."""

    def __init__(self, schema, *args, **kwargs):
        super(FlaskResourceSchema, self).__init__(schema, handle_http,
                                                  *args, **kwargs)


class FlaskRouter(Router):

    """A router class specifically for Flask.

    It provides default values that are common in Flask applications, while
    still allowing full customizability, if required.

    :see: :class:`~doctor.router.Router`
    """

    def __init__(self, schema_dir, resource_schema_class=None,
                 default_base_handler=None,
                 raise_response_validation_errors=False):
        if resource_schema_class is None:
            resource_schema_class = FlaskResourceSchema
        if default_base_handler is None:
            default_base_handler = flask_restful.Resource
        super(FlaskRouter, self).__init__(
            schema_dir, resource_schema_class, default_base_handler,
            raise_response_validation_errors)



import inspect
import os


class RouterException(Exception):
    """
    An exception specific to errors that can occur in the
    :class:`~doctor.router.Router` class.
    """
    pass


class Router(object):
    """A class that creates routes.

    This class handles dynamically generating handlers for HTTP methods and
    mapping the route to a logic function.

    :param str schema_dir: The absolute path to directory that contains the
        schema files.
    :param class resource_schema_class: The class to use for the
        :class:`~doctor.resource.ResourceSchema` instance.
    :param class default_base_handler: The default base handler that should
        be subclassed when dynamically creating the handlers from the routes
        and logic functions.  This will default to `object` if not specified.
    :param bool raise_response_validation_errors: True to raise errors
        for response validation exceptions, False to just log them
        and return gracefully.
    """

    #: This attribute is incremented every time a new handler class is
    #: generated to ensure uniquness in the handler class names.
    _num = 1

    def __init__(self, schema_dir, resource_schema_class,
                 default_base_handler=None,
                 raise_response_validation_errors=False):
        self.resource_schema_class = resource_schema_class
        self.schema_dir = schema_dir
        self.schemas = {}
        self.raise_response_validation_errors = raise_response_validation_errors
        if default_base_handler is not None:
            self.default_base_handler = default_base_handler
        else:
            self.default_base_handler = object

    def get_schema(self, schema_file):
        """Gets the resource schema for the provided file.

        :param str schema_file: The name of the schema file to load from
            :attr:`~schema_dir`.  e.g. 'app.yaml'
        :returns: An instance of `self.resource_schema_class`.
        """
        if self.schemas.get(schema_file):
            return self.schemas.get(schema_file)
        self.schemas[schema_file] = self.resource_schema_class.from_file(
            os.path.join(self.schema_dir, schema_file),
            self.raise_response_validation_errors)
        return self.schemas[schema_file]

    def _undecorate_func(self, func):
        """Returns the original function from the decorated one.

        The purpose of this function is to return the original `func` in the
        event that it has decorators attached to it, instead of the decorated
        function.

        :param function func: The function to unwrap.
        :returns: The unwrapped function.
        """
        while True:
            if func.__closure__:
                for cell in func.__closure__:
                    if inspect.isfunction(cell.cell_contents):
                        if func.__name__ == cell.cell_contents.__name__:
                            func = cell.cell_contents
                            break
            else:
                break
        return func

    def _get_params_from_func(self, func, omit_args=None):
        """Gets all parameters and required parameters from the func signature.

        This assumes all arguments of a function that don't have default
        values are required, and all others are optional.

        :param function func: The function to collect arguments from.
        :param list(str) omit_args: A list of function args to omit when
            examining the function signature.
        :returns: A tuple containing all parameters of the function signature
            and all required args.
        """
        argspec = inspect.getargspec(self._undecorate_func(func))
        if argspec.defaults:
            required = argspec.args[:-len(argspec.defaults)]
        else:
            required = argspec.args[:]
        params = argspec.args
        if omit_args:
            required = [r for r in required if r not in omit_args]
            params = [p for p in argspec.args if p not in omit_args]
        return (params, required)

    def create_routes(self, docs_group_title, schema_file, routes):
        """Creates a new set of routes.

        Routes will be returned as a list of tuples containing the route and
        handler class to map to.  e.g.

        [
            ('^/app/<app_id:int>/?$', AppHandler),
            ('^/app/?$', AppListHandler),
        ]

        :param str docs_group_title: The documentation title to group the
            routes underneath.
        :param str schema_file: The name of the schema file to use for
            request parameter and response definitions.  e.g. `app.yaml`
        :param dict routes: A dict of routes.  See :ref:`routes-docs` for the
            definition of the dict and possible options.
        :returns: A list of tuples containing the route and handler.  See
            example above.
        """
        schema = self.get_schema(schema_file)
        created_routes = []
        for route, methods in routes.iteritems():
            handler = None
            base_handler_class = methods.pop('base_handler_class',
                                             self.default_base_handler)
            # If a handler_name isn't defined, dynmaically generate one.  We
            # append a number to the Handler class name to ensure uniqueness.
            handler_name = methods.pop('handler_name',
                                       'Handler{}'.format(self._num))
            handler_decorators = methods.pop('decorators', [])
            handler_additional_args = methods.pop('additional_args', {})
            for method in methods:
                opts = routes[route][method]
                method = method.lower()
                if 'logic' not in opts:
                    raise RouterException(
                        '`logic` key not defined for {}: {}'.format(
                            method, opts))

                # Retrieve all params and required params from the logic
                # function signature.
                omit_args = opts.get('omit_args')
                params, required = self._get_params_from_func(
                    opts['logic'], omit_args)
                # If additiona_args was passed update params and required.
                # These are additional arguments that should be documented for
                # the request, but are not part of the logic function's
                # signature.
                additional_args = opts.get('additional_args', {})
                additional_args.update(handler_additional_args)
                if additional_args:
                    params.extend(additional_args.get('optional', []))
                    params.extend(additional_args.get('required', []))
                    required.extend(additional_args.get('required', []))

                # Check if the schema was overridden for the method.
                if opts.get('schema'):
                    method_schema = self.get_schema(opts['schema'])
                else:
                    method_schema = schema

                http_func = getattr(method_schema, 'http_' + method)
                func = http_func(
                    opts['logic'], params=params, required=required,
                    response=opts.get('response'), title=opts.get('title'),
                    allowed_exceptions=opts.get('allowed_exceptions'))
                # Apply all decoraters to the `func`
                decorators = opts.get('decorators', [])
                decorators.extend(handler_decorators)
                for decorator in decorators:
                    func = decorator(func)
                handler_methods_and_properites = {
                    '__name__': handler_name,
                    method: func,
                    'schematic_title': docs_group_title,
                }
                # If this route has more than one http method just add the
                # new method to the existing handler.
                if handler is not None:
                    setattr(handler, method,
                            handler_methods_and_properites[method])
                    # This is specific to Flask.  It's MethodView class
                    # initializes the methods attribute in __new__ so we
                    # need to add all the other http methods we are defining
                    # on the handler after it gets created by type.
                    if hasattr(handler, 'methods'):
                        handler.methods.append(method.upper())
                else:
                    # Dynamically generate a handler class with the http
                    # method attached.
                    handler = type(
                        handler_name, (base_handler_class,),
                        handler_methods_and_properites)
                    # Increment `_num` so that the next handler name created
                    # is different from the previous.
                    self._num += 1
                opts['logic']._handler = handler
            created_routes.append((route, handler))
        return created_routes



import os
import re
import sys

try:
    from sphinx.util.docstrings import prepare_docstring
except ImportError:
    prepare_docstring = None


#: Used to identify the end of the description block, and the beginning of the
#: parameters. This assumes that the parameters and such will always occur at
#: the end of the docstring.
DESCRIPTION_END_RE = re.compile(':(arg|param|returns|throws)', re.I)


def nested_set(data, keys, value):
    """Set a nested key value in a dict based on key list.

    :param data dict: the dict to set nested key in.
    :param keys list: the nested list of keys.
    :param value object: the final value to set.
    """
    for key in keys[:-1]:
        data = data.setdefault(key, {})
    data[keys[-1]] = value


def exec_params(call, *args, **kwargs):
    """Execute a callable with only the defined parameters
    and not just *args, **kwargs.

    :param callable call: The callable to exec with the given params as
        defined by itself. call should have an inspect.ArgSpec attached
        as an attribute _argspec
    :returns anything:
    :raises TypeError:
    """
    arg_spec = getattr(call, '_argspec', None)
    if arg_spec and not arg_spec.keywords:
        kwargs = {key: value for key, value in kwargs.iteritems()
                  if key in arg_spec.args}
    return call(*args, **kwargs)


def get_module_attr(module_filename, module_attr, namespace=None):
    """Get an attribute from a module.

    This uses execfile to load the module with a private namespace, and then
    plucks and returns the given attribute from that module's namespace.

    Note that, while this method doesn't have any explicit unit tests, it is
    tested implicitly by the doctor's own documentation. The Sphinx
    build process will fail to generate docs if this does not work.

    :param str module_filename: Path to the module to execute (e.g.
        "../src/app.py").
    :param str module_attr: Attribute to pluck from the module's namespace.
        (e.g. "app").
    :param dict namespace: Optional namespace. If one is not passed, an empty
        dict will be used instead. Note that this function mutates the passed
        namespace, so you can inspect a passed dict after calling this method
        to see how the module changed it.
    :returns: The attribute from the module.
    :raises KeyError: if the module doesn't have the given attribute.
    """
    if namespace is None:
        namespace = {}
    module_filename = os.path.abspath(module_filename)
    namespace['__file__'] = module_filename
    module_dir = os.path.dirname(module_filename)
    old_cwd = os.getcwd()
    old_sys_path = sys.path[:]
    try:
        os.chdir(module_dir)
        sys.path.append(module_dir)
        execfile(module_filename, namespace)
        return namespace[module_attr]
    finally:
        os.chdir(old_cwd)
        sys.path = old_sys_path


def get_description_lines(docstring):
    """Extract the description from the given docstring.

    This grabs everything up to the first occurrence of something that looks
    like a parameter description. The docstring will be dedented and cleaned
    up using the standard Sphinx methods.

    :param str docstring: The source docstring.
    :returns: list
    """
    if prepare_docstring is None:
        raise ImportError('sphinx must be installed to use this function.')

    if not isinstance(docstring, basestring):
        return []
    lines = []
    for line in prepare_docstring(docstring):
        if DESCRIPTION_END_RE.match(line):
            break
        lines.append(line)
    if lines and lines[-1] != '':
        lines.append('')
    return lines



import json

import mock
from sphinx.errors import SphinxError

from doctor.docs import base
from doctor.errors import SchemaError
from doctor.resource import ResourceSchema, ResourceSchemaAnnotation
from .base import TestCase


class TestDocsBase(TestCase):

    def setUp(self):
        super(TestDocsBase, self).setUp()
        self.schema = ResourceSchema({
            'definitions': {
                'a': {
                    'type': 'string',
                    'description': 'example description for a',
                    'example': 'example string',
                },
                'b': {
                    'type': 'integer',
                    'description': 'example description for b',
                    'example': 123,
                },
                'c': {
                    'type': 'boolean',
                    'description': 'example description for c',
                    'example': True,
                },
                'd': {
                    'type': ['integer', 'null'],
                    'description': 'example description for d',
                    'example': 456,
                },
                'e': {
                    'type': ['array'],
                    'description': 'example description for e',
                    'example': ['example', 'array'],
                },
                'f': {
                    'type': 'object',
                    'description': 'example description for f',
                    'properties': {
                        'str': {
                            'type': 'string',
                            'description': 'A string',
                            'example': 'A String',
                        },
                    },
                    'example': {'str': 'example string'},
                },
                'g': {
                    'type': 'string',
                    'description': 'example description for g',
                    'example': {
                        'oneOf': [
                            'example string 1',
                            'example string 2',
                        ],
                    },
                },
                'h': {
                    'type': ['null'],
                    'description': 'example description for h',
                    'example': None
                },
                'i': {
                    'type': 'array',
                    'description': 'example description for i',
                    'items': {
                        '$ref': '#/definitions/b',
                    },
                },
                # An example with an array of objects with no example
                # key on the object.
                'j': {
                    'type': 'array',
                    'description': 'An array of objects',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'str': {
                                'type': 'string',
                                'description': 'A String',
                                'example': 'string',
                            },
                            'int': {
                                'type': 'integer',
                                'description': 'An integer',
                                'example': 1022,
                            },
                        },
                        'additionalProperties': False,
                    },
                },
                'k': {
                    'type': 'string',
                    'description': 'The type of shirt.',
                    'enum': ['tshirt', 'sweater'],
                    'example': 'tshirt',
                },
                # An object w/o any properties.
                'l': {
                    'type': 'object',
                    'description': 'An object that can have any properties',
                    'example': {
                        'foo': 'bar',
                    },
                },
                # An object with a property that is a reference we can't resolve
                'm': {
                    'type': 'object',
                    'description': 'An object with a propertery that is a ref',
                    'example': {
                        'hey': 'ref',
                    },
                    'properties': {
                        'hey': {
                            '$ref': 'dinosaurs.yaml#/definitions/trex',
                        },
                    },
                },
                # An object with a property to document whose type is an
                # array, rather than a single item string.
                'n': {
                    'type': ['null', 'object'],
                    'description': 'example description for n',
                    'properties': {
                        'str': {
                            'type': 'string',
                            'description': 'A string',
                            'example': 'A String',
                        },
                    },
                    'example': {'str': 'example string `n`'},
                },
            },
        })
        self.request_schema = self.schema._create_request_schema(
            params=('a', 'b', 'c'), required=('a', 'b'))
        self.response_schema = {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'b': {'$ref': '#/definitions/b'},
                    'c': {'$ref': '#/definitions/c'},
                },
            },
        }

        def mock_logic():
            pass
        mock_logic.required_args = ['a', 'c']

        self.annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)

    def test_get_example_value(self):
        definitions = self.schema.resolve('#/definitions')
        results = [base.get_example_value(prop_schema, self.schema, key)
                   for key, prop_schema in sorted(definitions.iteritems())]
        expected = [
            'example string',
            123,
            True,
            456,
            ['example', 'array'],
            {'str': 'example string'},
            'example string 1',
            None,
            [123],
            [{'str': 'string', 'int': 1022}],
            'tshirt',
            {'foo': 'bar'},
            {'hey': 'ref'},
            {'str': 'example string `n`'},
        ]
        self.assertEqual(expected, results)

    def test_get_example_for_object(self):
        """
        This test verifies we can get an example for an object that is
        missing an example from normal properties.
        """
        schema = {
            'definitions': {
                'a': {
                    'type': 'object',
                    'properties': {
                        'foo': {
                            'type': 'string',
                            'description': 'A foo',
                            'example': 'Foo!',
                        },
                    },
                    'additionalProperties': False,
                },
            },
        }
        schema = ResourceSchema(schema)
        definitions = schema.resolve('#/definitions')
        result = base.get_example_for_object(definitions['a'], schema)
        expected = {'foo': 'Foo!'}
        self.assertEqual(expected, result)

    def test_get_example_for_object_with_different_schema(self):
        """
        This test verifies if the object_schema is a different file than the
        schema argument, that we create a new ResourceSchema instance from
        the object schema in order to resolve example values.
        """
        schema = ResourceSchema({
            'definitions': {
                'a': {
                    'type': 'object',
                    'properties': {
                        'foo': {
                            'type': 'string',
                            'description': 'A foo',
                            'example': 'Foo!',
                        },
                    },
                    'additionalProperties': False,
                },
            },
        })
        object_schema = {
            '$schema': 'http://json-schema.org/draft-04/schema#',
            'title': 'Another type of schema',
            'description': 'A description',
            'definitions': {
                'interesting_thing': {
                    'type': 'object',
                    'properties': {
                        'cat': {
                            'type': 'string',
                            'description': 'The name of a cat',
                            'example': 'Els',
                        },
                    },
                },
                'interesting_things': {
                    'type': 'array',
                    'description': 'an array of intersting things',
                    'items': {
                        '$ref': '#/definitions/interesting_thing',
                    },
                }
            },
            'type': 'object',
            'properties': {
                'interesting_things': {
                    '$ref': '#/definitions/interesting_things',
                },
            }
        }
        actual = base.get_example_for_object(object_schema, schema)
        expected = {'interesting_things': [{'cat': 'Els'}]}
        self.assertEqual(expected, actual)

    def test_get_example_value_anyof(self):
        """
        This test verifies that if a property has an anyOf declaration for the
        example, that we first resolve the anyOf by grabbing the first
        defintion in it's list.
        """
        property_schema = {
            'anyOf': [
                {'$ref': '#/definitions/b'},
                {'$ref': '#/definitions/c'}
            ]
        }
        actual = base.get_example_value(property_schema, self.schema, 'type')
        # expected value should be the example from definitions/b
        expected = 123
        self.assertEqual(expected, actual)

    def test_get_example_reraises_with_key(self):
        """
        This test verifies we re-raise the SchemaError with the offending key.
        """
        property_schema = {
            'type': 'integer',
            'description': 'An integer and we are missing the example'
        }
        with self.assertRaises(SchemaError) as cm:
            base.get_example_value(property_schema, self.schema, 'foobar')
        expected = ("Error resolving #example from {'type': 'integer', "
                    "'description': 'An integer and we are missing the "
                    "example'} for property `foobar`. Unresolvable JSON "
                    "pointer: u'example' (from )")
        self.assertEqual(expected, cm.exception.message)

    def test_get_example_lines_json(self):
        """Tests an example when the response is valid JSON."""
        headers = {'GeoIp-Country-Code': 'US'}
        lines = base.get_example_lines(headers, 'GET', 'http://example.com/',
                                       {}, json.dumps({'foo': 1, 'bar': 2}))
        self.assertEqual(lines, [
            '',
            'Example Request:',
            '',
            '.. code-block:: bash',
            '',
            '   curl http://example.com/ -X GET -H \'GeoIp-Country-Code: US\'',
            '',
            'Example Response:',
            '',
            '.. code-block:: json',
            '',
            '   {',
            '     "bar": 2,',
            '     "foo": 1',
            '   }',
        ])

    def test_get_example_lines_text(self):
        """Tests an example when the response is *not* valid JSON."""
        lines = base.get_example_lines({}, 'GET', 'http://example.com/', {},
                                       'hello, world!')
        self.assertEqual(lines, [
            '',
            'Example Request:',
            '',
            '.. code-block:: bash',
            '',
            '   curl http://example.com/ -X GET',
            '',
            'Example Response:',
            '',
            '.. code-block:: text',
            '',
            '   hello, world!',
        ])

    def test_get_json_object_lines(self):
        result = base.get_json_object_lines(
            self.annotation, self.request_schema, field='>json', url_params=[])
        self.assertEqual(result, [
            ':>json str a: example description for a',
            ':>json int b: example description for b',
            ':>json bool c: example description for c',
        ])

    def test_get_json_object_lines_for_request_with_enum(self):
        request_schema = self.schema._create_request_schema(
            params=('a', 'b', 'c', 'k'), required=('a', 'b'))
        result = base.get_json_object_lines(
            self.annotation, request_schema, field='>json', url_params=[],
            request=True)
        self.assertEqual(result, [
            ':>json str a: **Required**.  example description for a',
            ':>json bool c: **Required**.  example description for c',
            ':>json int b: example description for b',
            (":>json str k: The type of shirt. Must be one of: "
             "`['tshirt', 'sweater']`."),
        ])

    def test_get_json_object_lines_for_request(self):
        """
        This tests that when the request kwarg is True that any
        required params have the description prefixed with
        **Required** and sorted in alphabetical order, followed by
        any optional parameters in alpabetical order.
        """
        url_params = ['a']
        result = base.get_json_object_lines(
            self.annotation, self.request_schema, field='>json',
            url_params=url_params, request=True)
        self.assertEqual(result, [
            ':param str a: **Required**.  example description for a',
            ':>json bool c: **Required**.  example description for c',
            ':>json int b: example description for b',
        ])

    def test_get_json_object_lines_object(self):
        def mock_logic():
            pass
        mock_logic.required_args = ['f']
        request_schema = self.schema._create_request_schema(
            params=('f',), required=('f',))
        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)
        result = base.get_json_schema_lines(
            annotation, request_schema, field='>json', route='/foo')
        expected = [
            ':>json dict f|object: example description for f',
            ':>json str str|objectproperty: A string'
        ]
        self.assertEqual(expected, result)

    def test_get_json_object_lines_object_type_is_array(self):
        """
        This test verifies if a definition has a type of array and one of the
        types is an object that we still docuemnt the properties of the object.
        This is a regression test where we used to only document object
        properties if the type was a string equal to 'object'.
        """
        def mock_logic():
            pass
        mock_logic.required_args = ['n']
        request_schema = self.schema._create_request_schema(
            params=('n',), required=('n',))
        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)
        result = base.get_json_schema_lines(
            annotation, request_schema, field='>json', route='/foo')
        expected = [
            ':>json null,dict n|object: example description for n',
            ':>json str str|objectproperty: A string'
        ]
        self.assertEqual(expected, result)

    def test_get_json_object_lines_object_no_properties(self):
        """
        This is a regression test that ensures we don't blow up if we encounter
        an object that doesn't have any defined properties to document.
        """
        def mock_logic():
            pass
        mock_logic.required_args = ['l']
        request_schema = self.schema._create_request_schema(
            params=('l',), required=('l',))
        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)
        result = base.get_json_schema_lines(
            annotation, request_schema, field='>json', route='/foo')
        expected = [':>json dict l: An object that can have any properties']
        self.assertEqual(expected, result)

    def test_get_json_object_lines_object_property_is_a_ref(self):
        """
        This is a regression test that ensures we don't blow up if we encounter
        an object that has a property that can't be resolved.
        """
        def mock_logic():
            pass
        mock_logic.required_args = ['m']
        request_schema = self.schema._create_request_schema(
            params=('m',), required=('m',))
        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)
        result = base.get_json_schema_lines(
            annotation, request_schema, field='>json', route='/foo')
        expected = [':>json dict m: An object with a propertery that is a ref']
        self.assertEqual(expected, result)

    def test_get_json_schema_lines(self):
        result = base.get_json_schema_lines(
            self.annotation, self.request_schema, field='>json', route='/foo')
        self.assertEqual(result, [
            ':>json str a: example description for a',
            ':>json int b: example description for b',
            ':>json bool c: example description for c',
        ])

    def test_get_json_schema_lines_array(self):
        result = base.get_json_schema_lines(
            self.annotation, self.response_schema, field='<json', route='/foo')
        self.assertEqual(result, [
            ':<jsonarr int b: example description for b',
            ':<jsonarr bool c: example description for c',
        ])

    @mock.patch('doctor.docs.base.logging.warn')
    def test_get_json_schema_lines_array_invalid_type(self, mock_warn):
        self.response_schema['items']['type'] = 'string'
        result = base.get_json_schema_lines(
            self.annotation, self.response_schema, field='<json', route='/foo')
        self.assertEqual(result, [])
        self.assertEqual(mock_warn.call_args_list, [
            mock.call('Not documenting list item of type %s (for %s %s)',
                      'string', 'GET', '/foo')])

    @mock.patch('doctor.docs.base.logging.warn')
    def test_get_json_schema_lines_invalid_type(self, mock_warn):
        result = base.get_json_schema_lines(
            self.annotation, {'type': 'string'}, field='<json', route='/foo')
        self.assertEqual(result, [])
        self.assertEqual(mock_warn.call_args_list, [
            mock.call('Not documenting schema of type %s (for %s %s)',
                      'string', 'GET', '/foo')])

    def test_get_name(self):
        mock_class = mock.Mock()
        mock_class.__module__ = 'foo.bar'
        mock_class.__name__ = 'baz'
        self.assertEqual(base.get_name(mock_class), 'foo.bar.baz')
        mock_class.__module__ = '__builtin__'
        self.assertEqual(base.get_name(mock_class), 'baz')

    def test_resolve(self):
        definitions = self.schema.resolve('#/definitions')
        self.assertEqual(base.resolve('#a', definitions, self.annotation),
                         definitions['a'])
        with self.assertRaises(SphinxError):
            base.resolve('#invalid_ref', definitions, self.annotation)

    def test_resolve_with_oneof(self):
        """
        This test verifies that if a oneOf definition hits the resolve
        method, that we first resolve the oneOf by grabbing the first
        defintion in it's list.
        """
        fragment = 'type'
        definitions = {
            'oneOf': [
                {'$ref': '#/definitions/a'},
                {'$ref': '#/definitions/g'}
            ]
        }
        actual = base.resolve(fragment, definitions, self.annotation)

        expected = {
            'type': 'string',
            'description': 'example description for a',
            'example': 'example string',
        }
        self.assertEqual(expected, actual)

    def test_resolve_with_anyof(self):
        """
        This test verifies that if a anyOf definition hits the resolve
        method, that we first resolve the anyOf by grabbing the first
        defintion in it's list.
        """
        fragment = 'type'
        definitions = {
            'anyOf': [
                {'$ref': '#/definitions/b'},
                {'$ref': '#/definitions/c'}
            ]
        }
        actual = base.resolve(fragment, definitions, self.annotation)

        expected = {
            'type': 'integer',
            'description': 'example description for b',
            'example': 123,
        }
        self.assertEqual(expected, actual)

    def test_resolve_of_type(self):
        schema = {
            'oneOf': [
                {'$ref': '#/definitions/a'},
                {'$ref': '#/definitions/g'}
            ]
        }
        actual = base.resolve_of_type(self.annotation, schema)
        expected = {
            'type': 'string',
            'description': 'example description for a',
            'example': 'example string',
        }
        self.assertEqual(expected, actual)


class TestDocsBaseHarness(TestCase):

    def test_init(self):
        harness = base.BaseHarness('http://foo/')
        self.assertEqual(harness.url_prefix, 'http://foo')

    def test_get_annotation_heading_schematic_title(self):
        """
        This test verifies we use the schematic_title attribute of the
        handler if it is present.
        """
        handler = mock.Mock(schematic_title='Test Title')
        route = '^foo/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)
        self.assertEqual('Test Title', actual)

    def test_get_annotation_heading_class_path(self):
        """
        This test verifies if the class path has a resource name in it,
        that we use it for the heading.
        e.g. class <api.handlers.foo_bar.FooListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.foo_bar.FooBarListHandler>')
        route = '^foo_bar/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_generic_handlers(self):
        """
        This test verifies if our handlers are not in their own resource
        modules that we get the heading from the handler class name.
        e.g. class <api.handlers.handlers.FooBarListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.handlers.FooBarListHandler>')
        route = '^foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_class_name_only(self):
        """
        This test verifies that if our handler has no path and is just
        the class name that we get the heading from the name.
        e.g. class <FooBarListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class FooBarListHandler>')
        route = '^foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_class_path_internal(self):
        """
        This test verifies the path where the class path has the resource
        name in it and it's an internal route.
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.foo_bar.InternalFooBarListHandler>')
        route = '^internal/r/foo_bar/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar (Internal)'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_class_name_only_internal(self):
        """
        This test verifies the path where our handler has no path and just
        the class name and that class is internal.
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class InternalFooBarListHandler>')
        route = '^/internal/foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar (Internal)'
        self.assertEqual(expected, actual)
